#ifndef MYX_MATH_UNITS_HPP_
#define MYX_MATH_UNITS_HPP_

#include <type_traits>
#include <cmath>

namespace myx {

namespace math {

/*!
 * \brief radians_to_degrees  Перевод из радиан в градусы
 * \param radians             Значение в радианах
 * \return                    Значение в градусах
 */
template < typename T >
T radians_to_degrees( const T& radians, typename std::enable_if< std::is_floating_point< T >::value >::type* = 0 )
{
	static_assert( std::is_arithmetic< T >::value, "Arithmetic type is required." );
	return( radians * M_1_PI * 180.0 );
}


/*!
 * \brief degrees_to_radians  Перевод из градусов в радианы
 * \param degrees             Значение в градусах
 * \return                    Значение в радианах
 */
template < typename T >
T degrees_to_radians( const T& degrees, typename std::enable_if< std::is_floating_point< T >::value >::type* = 0 )
{
	static_assert( std::is_arithmetic< T >::value, "Arithmetic type is required." );
	return( degrees * M_PI / 180.0 );
}


/*!
 * \brief decibel_to_power  Перевод из децибелов в мощность $y = 10^{db / 10}$
 * \param db                Децибелы
 * \return                  Мощность
 */
template < typename T >
T decibel_to_power( T const& db )
{
	static_assert( std::is_arithmetic< T >::value, "Arithmetic type is required." );
	return( std::pow( 10, db / 10.0 ) );
}


/*!
 * \brief power_to_decibel  Перевод из мощности в децибелы по формуле $y = 10 * \log(x)$
 * \param power             Мощность
 * \return                  Децибелы
 */
template < typename T >
T power_to_decibel( T const& power )
{
	static_assert( std::is_arithmetic< T >::value, "Arithmetic type is required." );
	return( 10 * std::log( power ) );
}


/*!
 * \brief decibel_to_power  Перевод из децибелов в амплитуду $y = 10^{db / 20}$
 * \param db                Децибелы
 * \return                  Амплитуда
 */template < typename T >
T decibel_to_magnitude( T const& db )
{
	static_assert( std::is_arithmetic< T >::value, "Arithmetic type is required." );
	return( std::pow( 10, db / 20.0 ) );
}


/*!
 * \brief magnitude_to_decibel  Перевод из амплитуды в децибелы по формуле $y = 20 * \log(x)$
 * \param magnitude             Амплитуда
 * \return                      Децибелы
 */
template < typename T >
T magnitude_to_decibel( T const& magnitude )
{
	static_assert( std::is_arithmetic< T >::value, "Arithmetic type is required." );
	return( 20 * std::log( magnitude ) );
}

} // namespace math

} // namespace myx


#endif // MYX_MATH_UNITS_HPP_
